"""
Copyright 2021 Adobe
All Rights Reserved.

NOTICE: Adobe permits you to use, modify, and distribute this file in accordance
with the terms of the Adobe license agreement accompanying it.
"""


class KnownException(Exception):
    """
    Marker class for exceptions that should not have a full stacktrack printed out.
    """
